package com.javacodegeeks.SaxDomPerformance;

import java.io.*;

import java.util.List;
import java.util.regex.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.sax.SAXSource;
import javax.xml.xpath.*;

import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.TreeInfo;
import net.sf.saxon.trans.XPathException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;


public class SaxDomPerformance {

    // Default to the qual2017.xml file. Which can be downloaded
    // from the U.S. National Library of Medicine at the
    // following URL:
    //      https://www.nlm.nih.gov/mesh/download_mesh.html
    private String xml_file = "qual2017.xml";  // Actual XML file path

    private static final int TOTAL_TEST_COUNT = 5000;
    private static final String PROMPT_RUN_TIMED_TEST = "Run time test for " +
            TOTAL_TEST_COUNT + " executions? (Y or N)";

    private enum ParserType {DOM, SAX, None};
    private enum Mode {Eval, Test};
    private ParserType parserInUse = ParserType.None;

    private static final String PROMPT_XML = "Enter XML file name (abc.xml) "+
            "or leave blank to use ";
    private static final String PROMPT_USER_CMD = "Enter <expression>, !d, !s, !x, !q, !help: ";
    private static final String PROMPT_CMD_HELP =
            "!d to switch to DOM parser\n"+
            "!s to switch to the SAX parser\n"+
            "!x to change XML file\n\n"+
            "!q to exit (default)\n\n"+
            "!help displays this message\n\n"+
            "Any input that does not begin with \"!\""+
            " is assumed to be an expression.";
    private static final String PROMPT_DISPLAY_RESULTS = "Display results?  (Y or N)";
    private static final String PROMPT_DOM_PARSER = "Switching to DOM parser...";
    private static final String PROMPT_SAX_PARSER = "Switching to SAX parser...";
    private static final String PROMPT_DONE = "Done.";
    private static final String PROMPT_EXIT_CONFIRMATION = "Do you wish to exit?  (Y or N): ";
    private static final String PROMPT_NO_PARSER_SELECTED = "No parser selected. Please use:"+
            "!d (for DOM) or !s (for SAX)";
    private static final String PROMPT_XML_FILE = "XML file:";
    private static final String PROMPT_PARSER_TYPE = "Parser Type:";
    private static final String PROMPT_LAST_CMD = "Last command: ";
    private static final String PROMPT_AVG_TIME = "Average time to compile and evaluate: ";
    private static final String PROMPT_TOTAL = "Total:";

    private BufferedReader kbdReader = null; // Reads input from keyboard
    private String xpExpression = null; // String representation of XPath expression
    private String lastCmd = ""; // Record last command entered

    // DOM Parser and support structures
    private Document domDocument = null;
    private DocumentBuilder domParser = null;

    // SAX Parser and support structures
    private TreeInfo saxTreeInfo = null; // Provides context when evaluating SAX XPath

    private XPath xpathObj = null; // Compiles XPath expressions, used for both DOM and SAX

    // Method to prompt user for the XML file path.
    // Returns and sets xml_file member variable to user input
    private String promptUser(String prompt, String defaultValue) {
        System.out.print(prompt+"  ");
        if (defaultValue == null) {
            defaultValue = ""; // Avoid returning null
        }
        String userInput = null;
        try {
            userInput = kbdReader.readLine();
        } catch (IOException e) {
            e.printStackTrace();
        }
        if ((userInput == null) || (userInput.length() == 0)) {
            userInput = defaultValue; // Nothing entered, use default value
        }
        return userInput;
    }

    private DocumentBuilder getDOMParser() {
        // The Dom parser to be returned
        DocumentBuilder retParser = null;
        DocumentBuilderFactory builderFactory =
                DocumentBuilderFactory.newInstance();
        try {
            retParser = builderFactory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            e.printStackTrace();
            retParser = null;
        }
        return retParser;
    }

    // Initialize structures needed by the application
    private void initialize() {
        // Setup an InputStreamReader to read from the keyboard
        InputStreamReader reader = new InputStreamReader(System.in);
        kbdReader = new BufferedReader(reader);
    }

    // Get the XML file name
    private void getXMLFileName() {
        String localPrompt = PROMPT_XML + xml_file + ": ";
        xml_file = promptUser(localPrompt, xml_file);
    }

    // Initialize the DOM parser and load the DOM document
    private void initializeDOMParser() {
        System.out.print(PROMPT_DOM_PARSER);
        // Set SAX structures to null to release memory
        saxTreeInfo = null;
        xpathObj = null;

        // Instantiate the DOM parser
        domParser = getDOMParser();
        // Load the DOM Document from the XML data using the parser
        try {
            domDocument = domParser.parse(xml_file);
        } catch (SAXException | IOException e) {
            System.out.println("Exception in loadDOMDocument():  " + e.getMessage());
            e.printStackTrace();
        }
        // Instantiate an XPath object which compiles
        // and evaluates XPath expressions.
        xpathObj = XPathFactory.newInstance().newXPath();
        System.out.println(PROMPT_DONE);
    }

    // Initialize the SAX parser and it's support structures
    private void initializeSAXParser() {
        System.out.print(PROMPT_SAX_PARSER);
        // Set DOM structures to null to release memory
        domDocument = null;
        domParser = null;
        xpathObj = null;

        try {
            // The following initialization code is specific to Saxon
            // Please refer to SaxonHE documentation for details
            XPathFactory xpFactory = new net.sf.saxon.xpath.XPathFactoryImpl();
            xpathObj = xpFactory.newXPath();

            // Build the source document.
            InputSource inputSrc = new InputSource(new File(xml_file).getAbsolutePath());
            SAXSource saxSrc = new SAXSource(inputSrc);
            net.sf.saxon.Configuration config = ((net.sf.saxon.xpath.XPathFactoryImpl) xpFactory).getConfiguration();
            saxTreeInfo = config.buildDocumentTree(saxSrc);
        } catch (XPathException e) {
            System.out.println("Exception in initialize():  " + e.getMessage());
            e.printStackTrace();
        }
        System.out.println(PROMPT_DONE);
    }

    private void displayTimedResults(float testTimeTotal) {
        System.out.printf("%s %.4f (%s %.0f)\n", PROMPT_AVG_TIME,
                (float) (testTimeTotal / (float)TOTAL_TEST_COUNT),
                PROMPT_TOTAL, testTimeTotal);
    }

    // Display standardized numeric result
    private void displayNumericResult(String command,
                                      Number numResult,
                                      String fileName) {
        System.out.println(command + " counted " + numResult +
                " records in " + fileName + ".");
    }

    // Display standardized node contents
    private void displayNodeResult(int nodeIdx, String name,
                                      String value) {
        System.out.println((nodeIdx + 1) + ": " + name + " = " + value);    }

    // Execute XPath command based on selected parser
    private void executeXPath(String cmd, ParserType parserInUse,
                              boolean fNumberResult, Mode mode) {
        XPathExpression processCmd = null;
        Number resultNum = null;
        long testTimeStart = 0;
        // Setup the number of times to compile and evaluate
        // the expression based on the mode
        float testTimeTotal = 0;
        int maxTestCount = 1;
        if (mode == Mode.Test) {
            maxTestCount = TOTAL_TEST_COUNT;
        }
        try {
            switch (parserInUse) {
                case DOM:
                    if (fNumberResult) {
                        for (int testCount = 0; testCount < maxTestCount; testCount++) {
                            testTimeStart = System.currentTimeMillis();
                            processCmd = xpathObj.compile(cmd);
                            resultNum = (Number) processCmd.evaluate(domDocument, XPathConstants.NUMBER);
                            testTimeTotal += System.currentTimeMillis() - testTimeStart;
                        }
                        if (mode == Mode.Eval) {
                            displayNumericResult(cmd, resultNum, xml_file);
                        } else {
                            displayTimedResults(testTimeTotal);
                        }
                    } else {
                        // Result will be a list
                        NodeList resultNodeList = null;
                        for (int testCount = 0; testCount < maxTestCount; testCount++) {
                            testTimeStart = System.currentTimeMillis();
                            processCmd = xpathObj.compile(cmd);
                            resultNodeList = (NodeList) processCmd.evaluate(domDocument, XPathConstants.NODESET);
                            testTimeTotal += System.currentTimeMillis() - testTimeStart;
                        }
                        if (mode == Mode.Eval) {
                            if ((resultNodeList != null) && (resultNodeList.getLength() > 0)) {
                                int nodeCount = resultNodeList.getLength();
                                System.out.println("There are " + nodeCount + " nodes in the result:");

                                String displayResults = promptUser(PROMPT_DISPLAY_RESULTS, "N");
                                if (displayResults.equalsIgnoreCase("Y")) {
                                    // Go through each node in the list and display the node name and value
                                    for (int i = 0; i < nodeCount; i++) {
                                        Node node = resultNodeList.item(i);
                                        displayNodeResult(i, node.getNodeName(), node.getNodeValue());
                                    }
                                }
                            }
                        } else {
                            displayTimedResults(testTimeTotal);
                        }
                    }
                    break;
                case SAX:
                    if (fNumberResult) {
                        for (int testCount = 0; testCount < maxTestCount; testCount++) {
                            testTimeStart = System.currentTimeMillis();
                            processCmd = xpathObj.compile(cmd);
                            resultNum = (Number) processCmd.evaluate(saxTreeInfo, XPathConstants.NUMBER);
                            testTimeTotal += System.currentTimeMillis() - testTimeStart;
                        }
                        if (mode == Mode.Eval) {
                            displayNumericResult(cmd, resultNum, xml_file);
                        } else {
                            displayTimedResults(testTimeTotal);
                        }
                    } else {
                        // Result will be a list of nodes
                        List resultNodeList = null;
                        for (int testCount = 0; testCount < maxTestCount; testCount++) {
                            testTimeStart = System.currentTimeMillis();
                            processCmd = xpathObj.compile(cmd);
                            resultNodeList = (List) processCmd.evaluate(saxTreeInfo, XPathConstants.NODESET);
                            testTimeTotal += System.currentTimeMillis() - testTimeStart;
                        }
                        if (mode == Mode.Eval) {
                            if ((resultNodeList != null) && (resultNodeList.size() > 0)) {
                                int nodeCount = resultNodeList.size();
                                System.out.println("There are " + nodeCount + " nodes in the result.");

                                String displayResults = promptUser(PROMPT_DISPLAY_RESULTS, "N");
                                if (displayResults.equalsIgnoreCase("Y")) {
                                    // Go through each node in the list and display the node name and value
                                    for (int i = 0; i < nodeCount; i++) {
                                        NodeInfo cNode = (NodeInfo) resultNodeList.get(i);
                                        displayNodeResult(i, cNode.getDisplayName(),  cNode.getStringValue());
                                    }
                                }
                            }
                        } else {
                            displayTimedResults(testTimeTotal);
                        }
                    }
                    break;
                default:
                    return;
            }
        } catch (XPathExpressionException e) {
            System.out.println("Exception in executeXPath():  " + e.getMessage());
            e.printStackTrace();
        }
    }

    // Show current status of the system
    private void displayStatus() {
        System.out.println("=+=+=+=+=+=+=+=+");
        System.out.println(PROMPT_XML_FILE + xml_file);
        System.out.println(PROMPT_PARSER_TYPE + parserInUse.toString());
        System.out.println(PROMPT_LAST_CMD + lastCmd);
        System.out.println("\n");
    }

    // First initialize the parsers, then loop asking which parser
    // to use and expression to be evaluated
    private void run() {
        initialize();
        String cmd = null;
        while (true) {
            displayStatus();
            cmd = promptUser(PROMPT_USER_CMD, "!q"); // Default to exit.
            if (cmd.equalsIgnoreCase("!x")) {
                // Change XML file
                getXMLFileName();
            } else if (cmd.equalsIgnoreCase("!d")) {
                // Select DOM parser
                initializeDOMParser();
                parserInUse = ParserType.DOM;
            } else if (cmd.equalsIgnoreCase("!s")) {
                // Select SAX parser
                initializeSAXParser();
                parserInUse = ParserType.SAX;
            } else if (cmd.equalsIgnoreCase("!q")) {
                // Ask if user really wants to quit
                cmd = promptUser(PROMPT_EXIT_CONFIRMATION, "N");
                if (cmd.equalsIgnoreCase("Y")) {
                    System.exit(0); // Bye!
                }
            } else if (cmd.equalsIgnoreCase("!help")) {
                System.out.println(PROMPT_CMD_HELP);
            } else {
                // Assume xpath command was entered
                lastCmd = cmd; // Record to display in displayStatus()
                if (parserInUse == ParserType.None) {
                    System.out.println(PROMPT_NO_PARSER_SELECTED);
                } else {
                    // If the cmd matches the regex "count(*)" then the result will be a number
                    // else it will be a list.
                    boolean fNumberResult = Pattern.matches("count(.*)",cmd);
                    executeXPath(cmd, parserInUse, fNumberResult, Mode.Eval);
                    cmd = promptUser(PROMPT_RUN_TIMED_TEST, "y");
                    if (cmd.equalsIgnoreCase("Y")) {
                        executeXPath(cmd, parserInUse, fNumberResult, Mode.Test);
                    }
                }
            }
        }  // end while
    }


    public static void main(String[] args) {
        SaxDomPerformance appObj = new SaxDomPerformance();
        appObj.run();
    }

}
